###############################################################################
# Name: outbuff.py                                                            #
# Purpose: Gui and helper classes for running processes and displaying output #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Editra Control Library: OutputBuffer

This module contains classes that are useful for displaying output from running
tasks and processes. The classes are divided into three main categories, gui
classes, mixins, and thread classes. All the classes can be used together to
easily create multithreaded gui display classes without needing to worry about
the details and thread safety of the gui.

For example usage of these classes see ed_log and the Editra's Launch plugin

Class OutputBuffer:
This is the main class exported by this module. It provides a readonly output
display buffer that when used with the other classes in this module provides an
easy way to display continuous output from other processes and threads. It
provides two methods for subclasses to override if they wish to perform custom
handling.

  - Override the ApplyStyles method to do any processing and coloring of the
    text as it is put in the buffer.
  - Override the DoHotSpotClicked method to handle any actions to take when a
    hotspot has been clicked in the buffer.
  - Override the DoUpdatesEmpty method to perform any idle processing when no
    new text is waiting to be processed.

Class ProcessBufferMixin:
Mixin class for the L{OutputBuffer} class that provides handling for when an
OutputBuffer is used with a L{ProcessThread}. It provides three methods that can
be overridden in subclasses to perform extra processing.

  - DoProcessStart: Called as the process is being started in the ProcessThread,
                    it receives the process command string as an argument.
  - DoFilterInput: Called as each chunk of output comes from the running process
                   use it to filter the results before displaying them in the
                   buffer.
  - DoProcessExit: Called when the running process has exited. It receives the
                   processes exit code as a parameter.

Class ProcessThread:
Thread class for running subprocesses and posting the output to an
L{OutputBuffer} via events.

Class TaskThread:
Thread class for running a callable. For optimal performance and responsiveness
the callable should be a generator object. All results are directed to an
L{OutputBuffer} through its AppendUpdate method.

Requirements:
  * wxPython 2.8
  * Macintosh/Linux/Unix Python 2.4+
  * Windows Python 2.5+ (ctypes is needed)

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: outbuff.py 69743 2011-11-12 20:22:48Z CJP $"
__revision__ = "$Revision: 69743 $"

__all__ = ["OutputBuffer", "OutputBufferEvent", "ProcessBufferMixin",
           "ProcessThreadBase", "ProcessThread", "TaskThread", "TaskObject",
           "OPB_STYLE_DEFAULT", "OPB_STYLE_INFO",
           "OPB_STYLE_WARN", "OPB_STYLE_ERROR", "OPB_STYLE_MAX",

           "OPB_ERROR_NONE", "OPB_ERROR_INVALID_COMMAND",

           "edEVT_PROCESS_START", "EVT_PROCESS_START", "edEVT_TASK_START",
           "EVT_TASK_START", "edEVT_UPDATE_TEXT", "EVT_UPDATE_TEXT",
           "edEVT_PROCESS_EXIT", "EVT_PROCESS_EXIT", "edEVT_TASK_COMPLETE",
           "EVT_TASK_COMPLETE", "edEVT_PROCESS_ERROR", "EVT_PROCESS_ERROR"]

#--------------------------------------------------------------------------#
# Imports
import os
import sys
import time
import errno
import signal
import threading
import types
import subprocess
import wx
import wx.stc

# Platform specific modules needed for killing processes
if subprocess.mswindows:
    import msvcrt
    import ctypes
else:
    import shlex
    import select
    import fcntl

#--------------------------------------------------------------------------#
# Globals
OUTPUTBUFF_NAME_STR = u'EditraOutputBuffer'
THREADEDBUFF_NAME_STR = u'EditraThreadedBuffer'

# Style Codes
OPB_STYLE_DEFAULT = 0  # Default Black text styling
OPB_STYLE_INFO    = 1  # Default Blue text styling
OPB_STYLE_WARN    = 2  # Default Red text styling
OPB_STYLE_ERROR   = 3  # Default Red/Hotspot text styling
OPB_STYLE_MAX     = 3  # Highest style byte used by outputbuffer

# All Styles
OPB_ALL_STYLES = (wx.stc.STC_STYLE_DEFAULT, wx.stc.STC_STYLE_CONTROLCHAR,
                  OPB_STYLE_DEFAULT, OPB_STYLE_ERROR, OPB_STYLE_INFO,
                  OPB_STYLE_WARN)

# Error Codes
OPB_ERROR_NONE            = 0
OPB_ERROR_INVALID_COMMAND = -1

#--------------------------------------------------------------------------#

# Event for notifying that the process has started running
# GetValue will return the command line string that started the process
edEVT_PROCESS_START = wx.NewEventType()
EVT_PROCESS_START = wx.PyEventBinder(edEVT_PROCESS_START, 1)

# Event for notifying that a task is starting to run
edEVT_TASK_START = wx.NewEventType()
EVT_TASK_START = wx.PyEventBinder(edEVT_TASK_START, 1)

# Event for passing output data to buffer
# GetValue returns the output text retrieved from the process
edEVT_UPDATE_TEXT = wx.NewEventType()
EVT_UPDATE_TEXT = wx.PyEventBinder(edEVT_UPDATE_TEXT, 1)

# Event for notifying that the the process has finished and no more update
# events will be sent. GetValue will return the processes exit code
edEVT_PROCESS_EXIT = wx.NewEventType()
EVT_PROCESS_EXIT = wx.PyEventBinder(edEVT_PROCESS_EXIT, 1)

# Event to notify that a process has completed
edEVT_TASK_COMPLETE = wx.NewEventType()
EVT_TASK_COMPLETE = wx.PyEventBinder(edEVT_TASK_COMPLETE, 1)

# Event to notify that an error occurred in the process
edEVT_PROCESS_ERROR = wx.NewEventType()
EVT_PROCESS_ERROR = wx.PyEventBinder(edEVT_PROCESS_ERROR, 1)

class OutputBufferEvent(wx.PyCommandEvent):
    """Event for data transfer and signaling actions in the L{OutputBuffer}"""
    def __init__(self, etype, eid=wx.ID_ANY, value=''):
        """Creates the event object"""
        super(OutputBufferEvent, self).__init__(etype, eid)

        # Attributes
        self._value = value
        self._errmsg = None

    #---- Properties ----#
    Value = property(lambda self: self.GetValue(),
                     lambda self, v: setattr(self, '_value', v))
    ErrorMessage = property(lambda self: self.GetErrorMessage(),
                            lambda self, msg: self.SetErrorMessage(msg))

    def GetValue(self):
        """Returns the value from the event.
        @return: the value of this event

        """
        return self._value

    def GetErrorMessage(self):
        """Get the error message value
        @return: Exception traceback string or None

        """
        return self._errmsg

    def SetErrorMessage(self, msg):
        """Set the error message value
        @param msg: Exception traceback string

        """
        try:
            tmsg = unicode(msg)
        except:
            tmsg = None
        self._errmsg = msg

#--------------------------------------------------------------------------#

class OutputBuffer(wx.stc.StyledTextCtrl):
    """OutputBuffer is a general purpose output display for showing text. It
    provides an easy interface for the buffer to interact with multiple threads
    that may all be sending updates to the buffer at the same time. Methods for
    styling and filtering output are also available.

    """
    def __init__(self, parent, id=wx.ID_ANY,
                 pos=wx.DefaultPosition,
                 size=wx.DefaultSize,
                 style=wx.BORDER_SUNKEN,
                 name=OUTPUTBUFF_NAME_STR):
        super(OutputBuffer, self).__init__(parent, id, pos,
                                           size, style, name)

        # Attributes
        self._mutex = threading.Lock()
        self._updating = threading.Condition(self._mutex)
        self._updates = list()
        self._timer = wx.Timer(self)
        self._line_buffer = -1                
        self._colors = dict(defaultb=(255, 255, 255), defaultf=(0, 0, 0),
                            errorb=(255, 255, 255), errorf=(255, 0, 0),
                            infob=(255, 255, 255), infof=(0, 0, 255),
                            warnb=(255, 255, 255), warnf=(255, 0, 0))

        # Setup
        self.__ConfigureSTC()

        # Event Handlers
        self.Bind(wx.EVT_TIMER, self.OnTimer)
        self.Bind(wx.stc.EVT_STC_HOTSPOT_CLICK, self._OnHotSpot)

    def __del__(self):
        """Ensure timer is cleaned up when we are deleted"""
        if self._timer.IsRunning():
            self._timer.Stop()

    def __ConfigureSTC(self):
        """Setup the stc to behave/appear as we want it to
        and define all styles used for giving the output context.
        @todo: make more of this configurable

        """
        self.SetMargins(3, 3)
        self.SetMarginWidth(0, 0)
        self.SetMarginWidth(1, 0)

        # To improve performance at cost of memory cache the document layout
        self.SetLayoutCache(wx.stc.STC_CACHE_DOCUMENT)
        self.SetUndoCollection(False) # Don't keep undo history
        self.SetReadOnly(True)
        self.SetCaretWidth(0)

        if wx.Platform == '__WXMSW__':
            self.SetEOLMode(wx.stc.STC_EOL_CRLF)
        else:
            self.SetEOLMode(wx.stc.STC_EOL_LF)

        #self.SetEndAtLastLine(False)
        self.SetVisiblePolicy(1, wx.stc.STC_VISIBLE_STRICT)

        # Define Styles
        highlight = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        self.SetSelBackground(True, highlight)
        if sum(highlight.Get()) < 384:
            self.SetSelForeground(True, wx.WHITE)
        else:
            self.SetSelForeground(True, wx.BLACK)
        self.__SetupStyles()

    def FlushBuffer(self):
        """Flush the update buffer
        @postcondition: The update buffer is empty

        """
        self._updating.acquire()
        self.SetReadOnly(False)
        txt = u''.join(self._updates[:])
        start = self.GetLength()
        if u'\0' in txt:
            # HACK: handle displaying NULLs in the STC
            self.AddStyledText('\0'.join(txt.encode('utf-8'))+'\0')
        else:
            self.AppendText(txt)
        self.GotoPos(self.GetLength())
        self._updates = list()
        self.ApplyStyles(start, txt)
        self.SetReadOnly(True)
        self.RefreshBufferedLines()
        self._updating.release()

    def __SetupStyles(self, font=None):
        """Setup the default styles of the text in the buffer
        @keyword font: wx.Font to use or None to use default

        """
        if font is None:
            if wx.Platform == '__WXMAC__':
                fsize = 11
            else:
                fsize = 10

            font = wx.Font(fsize, wx.FONTFAMILY_MODERN,
                           wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL)
        style = (font.GetFaceName(), font.GetPointSize(), "#FFFFFF")
        wx.stc.StyledTextCtrl.SetFont(self, font)

        # Custom Styles
        self.StyleSetSpec(OPB_STYLE_DEFAULT,
                          "face:%s,size:%d,fore:#000000,back:%s" % style)
        self.StyleSetSpec(OPB_STYLE_INFO,
                          "face:%s,size:%d,fore:#0000FF,back:%s" % style)
        self.StyleSetSpec(OPB_STYLE_WARN,
                          "face:%s,size:%d,fore:#FF0000,back:%s" % style)
        self.StyleSetSpec(OPB_STYLE_ERROR,
                          "face:%s,size:%d,fore:#FF0000,back:%s" % style)
        self.StyleSetHotSpot(OPB_STYLE_ERROR, True)

        # Default Styles
        self.StyleSetSpec(wx.stc.STC_STYLE_DEFAULT, \
                          "face:%s,size:%d,fore:#000000,back:%s" % style)
        self.StyleSetSpec(wx.stc.STC_STYLE_CONTROLCHAR, \
                          "face:%s,size:%d,fore:#000000,back:%s" % style)
        self.Colourise(0, -1)

    def _OnHotSpot(self, evt):
        """Handle hotspot clicks"""
        pos = evt.GetPosition()
        self.DoHotSpotClicked(pos, self.LineFromPosition(pos))

    #---- Public Member Functions ----#

    def AppendUpdate(self, value):
        """Buffer output before adding to window. This method can safely be
        called from non gui threads to add updates to the buffer, that will
        be displayed during the next idle period.
        @param value: update string to append to stack

        """
        self._updating.acquire()
        if not (type(value) is types.UnicodeType):
            value = value.decode(sys.getfilesystemencoding())
        self._updates.append(value)
        self._updating.release()

    def ApplyStyles(self, start, txt):
        """Apply coloring to text starting at start position.
        Override this function to do perform any styling that you want
        done on the text.
        @param start: Start position of text that needs styling in the buffer
        @param txt: The string of text that starts at the start position in the
                    buffer.

        """
        pass

    def CanCopy(self):
        """Is it possible to copy text right now
        @return: bool

        """
        sel = self.GetSelection()
        return sel[0] != sel[1]

    def CanCut(self):
        """Is it possible to Cut
        @return: bool

        """
        return not self.GetReadOnly()

    def Clear(self):
        """Clear the Buffer"""
        self.SetReadOnly(False)
        self.ClearAll()
        self.EmptyUndoBuffer()
        self.SetReadOnly(True)

    def DoHotSpotClicked(self, pos, line):
        """Action to perform when a hotspot region is clicked in the buffer.
        Override this function to provide handling of hotspots.
        @param pos: Position in buffer of where the click occurred.
        @param line: Line in which the click occurred (zero based index)

        """
        pass

    def DoUpdatesEmpty(self):
        """Called when update stack is empty
        Override this function to perform actions when there are no updates
        to process. It can be used for things such as temporarily stopping
        the timer or performing idle processing.

        """
        pass

    def GetDefaultBackground(self):
        """Get the default text style background color
        @return: wx.Colour

        """
        return wx.Colour(*self._colors['defaultb'])

    def GetDefaultForeground(self):
        """Get the default text style foreground color
        @return: wx.Colour

        """
        return wx.Colour(*self._colors['defaultf'])

    def GetErrorBackground(self):
        """Get the error text style background color
        @return: wx.Colour

        """
        return wx.Colour(*self._colors['errorb'])

    def GetErrorForeground(self):
        """Get the error text style foreground color
        @return: wx.Colour

        """
        return wx.Colour(*self._colors['errorf'])

    def GetInfoBackground(self):
        """Get the info text style background color
        @return: wx.Colour

        """
        return wx.Colour(*self._colors['infob'])

    def GetInfoForeground(self):
        """Get the info text style foreground color
        @return: wx.Colour

        """
        return wx.Colour(*self._colors['infof'])

    def GetWarningBackground(self):
        """Get the warning text style background color
        @return: wx.Colour

        """
        return wx.Colour(*self._colors['warnb'])

    def GetWarningForeground(self):
        """Get the warning text style foreground color
        @return: wx.Colour

        """
        return wx.Colour(*self._colors['warnf'])

    def GetUpdateQueue(self):
        """Gets a copy of the current update queue"""
        self._updating.acquire()
        val = list(self._updates)
        self._updating.release()
        return val

    def IsRunning(self):
        """Return whether the buffer is running and ready for output
        @return: bool

        """
        return self._timer.IsRunning()

    def OnTimer(self, evt):
        """Process and display text from the update buffer
        @note: this gets called many times while running thus needs to
               return quickly to avoid blocking the ui.

        """
        if len(self._updates):
            self.FlushBuffer()
        elif evt is not None:
            self.DoUpdatesEmpty()
        else:
            pass

    def RefreshBufferedLines(self):
        """Refresh and readjust the lines in the buffer to fit the current
        line buffering limits.
        @postcondition: Oldest lines are removed until we are back within the
                        buffer limit bounds.

        """
        if self._line_buffer < 0:
            return

        self.SetReadOnly(False)
        while self.GetLineCount() > self._line_buffer:
            self.SetCurrentPos(0)
            self.LineDelete()
        self.SetReadOnly(True)
        self.SetCurrentPos(self.GetLength())

    def SetDefaultColor(self, fore=None, back=None):
        """Set the colors for the default text style
        @keyword fore: Foreground Color
        @keyword back: Background Color

        """
        if fore is not None:
            self.StyleSetForeground(wx.stc.STC_STYLE_DEFAULT, fore)
            self.StyleSetForeground(wx.stc.STC_STYLE_CONTROLCHAR, fore)
            self.StyleSetForeground(OPB_STYLE_DEFAULT, fore)
            self._colors['defaultf'] = fore.Get()

        if back is not None:
            self.StyleSetBackground(wx.stc.STC_STYLE_DEFAULT, back)
            self.StyleSetBackground(wx.stc.STC_STYLE_CONTROLCHAR, back)
            self.StyleSetBackground(OPB_STYLE_DEFAULT, back)
            self._colors['defaultb'] = back.Get()

    def SetErrorColor(self, fore=None, back=None):
        """Set color for error text
        @keyword fore: Foreground Color
        @keyword back: Background Color

        """
        if fore is not None:
            self.StyleSetForeground(OPB_STYLE_ERROR, fore)
            self._colors['errorf'] = fore.Get()

        if back is not None:
            self.StyleSetBackground(OPB_STYLE_ERROR, back)
            self._colors['errorb'] = back.Get()

    def SetInfoColor(self, fore=None, back=None):
        """Set color for info text
        @keyword fore: Foreground Color
        @keyword back: Background Color

        """
        if fore is not None:
            self.StyleSetForeground(OPB_STYLE_INFO, fore)
            self._colors['infof'] = fore.Get()

        if back is not None:
            self.StyleSetBackground(OPB_STYLE_INFO, back)
            self._colors['infob'] = back.Get()

    def SetWarningColor(self, fore=None, back=None):
        """Set color for warning text
        @keyword fore: Foreground Color
        @keyword back: Background Color

        """
        if fore is not None:
            self.StyleSetForeground(OPB_STYLE_WARN, fore)
            self._colors['warnf'] = fore.Get()

        if back is not None:
            self.StyleSetBackground(OPB_STYLE_WARN, back)
            self._colors['warnb'] = back.Get()

    def SetFont(self, font):
        """Set the font used by all text in the buffer
        @param font: wxFont

        """
        for style in OPB_ALL_STYLES:
            self.StyleSetFont(style, font)

    def SetLineBuffering(self, num):
        """Set how many lines the buffer should keep for display.
        @param num: int (-1 == unlimited)

        """
        self._line_buffer = num
        self.RefreshBufferedLines()

    def SetText(self, text):
        """Set the text that is shown in the buffer
        @param text: text string to set as buffers current value

        """
        self.SetReadOnly(False)
        wx.stc.StyledTextCtrl.SetText(self, text)
        self.SetReadOnly(True)

    def Start(self, interval):
        """Start the window's timer to check for updates
        @param interval: interval in milliseconds to do updates

        """
        self._timer.Start(interval)

    def Stop(self):
        """Stop the update process of the buffer"""
        # Dump any output still left in tmp buffer before stopping
        self.OnTimer(None)
        self._timer.Stop()
        self.SetReadOnly(True)

#-----------------------------------------------------------------------------#

class ProcessBufferMixin:
    """Mixin class for L{OutputBuffer} to handle events
    generated by a L{ProcessThread}.

    """
    def __init__(self, update=100):
        """Initialize the mixin
        @keyword update: The update interval speed in msec

        """
        # Attributes
        self._rate = update

        # Event Handlers
        self.Bind(EVT_PROCESS_START, self._OnProcessStart)
        self.Bind(EVT_UPDATE_TEXT, self._OnProcessUpdate)
        self.Bind(EVT_PROCESS_EXIT, self._OnProcessExit)
        self.Bind(EVT_PROCESS_ERROR, self._OnProcessError)

    def _OnProcessError(self, evt):
        """Handle EVT_PROCESS_ERROR"""
        self.DoProcessError(evt.GetValue(), evt.GetErrorMessage())

    def _OnProcessExit(self, evt):
        """Handles EVT_PROCESS_EXIT"""
        self.DoProcessExit(evt.GetValue())

    def _OnProcessStart(self, evt):
        """Handles EVT_PROCESS_START"""
        self.DoProcessStart(evt.GetValue())
        self.Start(self._rate)

    def _OnProcessUpdate(self, evt):
        """Handles EVT_UPDATE_TEXT"""
        txt = self.DoFilterInput(evt.GetValue())
        self.AppendUpdate(txt)

    def DoFilterInput(self, txt):
        """Override this method to do an filtering on input that is sent to
        the buffer from the process text. The return text is what is put in
        the buffer.
        @param txt: incoming update text
        @return: string

        """
        return txt

    def DoProcessError(self, code, excdata=None):
        """Override this method to do any ui notification of when errors happen
        in running the process.
        @param code: an OBP error code
        @keyword excdata: Exception Data from process error
        @return: None

        """
        pass

    def DoProcessExit(self, code=0):
        """Override this method to do any post processing after the running
        task has exited. Typically this is a good place to call
        L{OutputBuffer.Stop} to stop the buffers timer.
        @keyword code: Exit code of program
        @return: None

        """
        self.Stop()

    def DoProcessStart(self, cmd=''):
        """Override this method to do any pre-processing before starting
        a processes output.
        @keyword cmd: Command used to start program
        @return: None

        """
        pass

    def SetUpdateInterval(self, value):
        """Set the rate at which the buffer outputs update messages. Set to
        a higher number if the process outputs large amounts of text at a very
        high rate.
        @param value: rate in milliseconds to do updates on

        """
        self._rate = value

#-----------------------------------------------------------------------------#

class ProcessThreadBase(threading.Thread):
    """Base Process Thread
    Override DoPopen in subclasses.

    """
    def __init__(self, parent):
        super(ProcessThreadBase, self).__init__()

        # Attributes
        self.abort = False          # Abort Process
        self._proc = None
        self._parent = parent       # Parent Window/Event Handler
        self._sig_abort = signal.SIGTERM    # default signal to kill process
        self._last_cmd = u""        # Last run command

    #---- Properties ----#
    LastCommand = property(lambda self: self._last_cmd,
                           lambda self, val: setattr(self, '_last_cmd', val)) 
    Parent = property(lambda self: self._parent)
    Process = property(lambda self: self._proc)

    def __DoOneRead(self):
        """Read one line of output and post results.
        @return: bool (True if more), (False if not)

        """
        if subprocess.mswindows:
            # Windows nonblocking pipe read implementation
            read = u''
            try:
                handle = msvcrt.get_osfhandle(self._proc.stdout.fileno())
                avail = ctypes.c_long()
                ctypes.windll.kernel32.PeekNamedPipe(handle, None, 0, 0,
                                                     ctypes.byref(avail), None)
                if avail.value > 0:
                    read = self._proc.stdout.read(avail.value)
                    if read.endswith(os.linesep):
                        read = read[:-1 * len(os.linesep)]
                else:
                    if self._proc.poll() is None:
                        time.sleep(1)
                        return True
                    else:
                        # Process has Exited
                        return False
            except ValueError, msg:
                return False
            except (subprocess.pywintypes.error, Exception), msg:
                if msg[0] in (109, errno.ESHUTDOWN):
                    return False
        else:
            # OSX and Unix nonblocking pipe read implementation
            if self._proc.stdout is None:
                return False

            flags = fcntl.fcntl(self._proc.stdout, fcntl.F_GETFL)
            if not self._proc.stdout.closed:
                fcntl.fcntl(self._proc.stdout,
                            fcntl.F_SETFL,
                            flags|os.O_NONBLOCK)

            try:
                try:
                    if not select.select([self._proc.stdout], [], [], 1)[0]:
                        return True

                    read = self._proc.stdout.read(4096)
                    if read == '':
                        return False
                except IOError, msg:
                    return False
            finally:
                if not self._proc.stdout.closed:
                    fcntl.fcntl(self._proc.stdout, fcntl.F_SETFL, flags)

        # Ignore encoding errors and return an empty line instead
        try:
            result = read.decode(sys.getfilesystemencoding())
        except UnicodeDecodeError:
            result = os.linesep

        if self.Parent:
            evt = OutputBufferEvent(edEVT_UPDATE_TEXT, self.Parent.GetId(), result)
            wx.PostEvent(self.Parent, evt)
            return True
        else:
            return False # Parent is dead no need to keep running

    def __KillPid(self, pid):
        """Kill a process by process id, causing the run loop to exit
        @param pid: Id of process to kill

        """
        # Dont kill if the process if it is the same one we
        # are running under (i.e we are running a shell command)
        if pid == os.getpid():
            return

        if wx.Platform != '__WXMSW__':
            # Close output pipe(s)
            try:
                try:
                    self._proc.stdout.close()
                except Exception, msg:
                    pass
            finally:
                self._proc.stdout = None

            # Try to kill the group
            try:
                os.kill(pid, self._sig_abort)
            except OSError, msg:
                pass

            # If still alive shoot it again
            if self._proc.poll() is not None:
                try:
                    os.kill(-pid, signal.SIGKILL)
                except OSError, msg:
                    pass

            # Try and wait for it to cleanup
            try:
                os.waitpid(pid, os.WNOHANG)
            except OSError, msg:
                pass

        else:
            # 1 == PROCESS_TERMINATE
            handle = ctypes.windll.kernel32.OpenProcess(1, False, pid)
            ctypes.windll.kernel32.TerminateProcess(handle, -1)
            ctypes.windll.kernel32.CloseHandle(handle)

    #---- Public Member Functions ----#
    def Abort(self, sig=signal.SIGTERM):
        """Abort the running process and return control to the main thread"""
        self._sig_abort = sig
        self.abort = True

    def DoPopen(self):
        """Open the process
        Override in a subclass to implement custom process opening
        @return: subprocess.Popen instance

        """
        raise NotImplementedError("Must implement DoPopen in subclasses!")

    def run(self):
        """Run the process until finished or aborted. Don't call this
        directly instead call self.start() to start the thread else this will
        run in the context of the current thread.
        @note: overridden from Thread

        """
        err = None
        try:
            self._proc = self.DoPopen()
        except OSError, msg:
            # NOTE: throws WindowsError on Windows which is a subclass of
            #       OSError, so it will still get caught here.
            if self.Parent:
                err =  OutputBufferEvent(edEVT_PROCESS_ERROR,
                                         self.Parent.GetId(),
                                         OPB_ERROR_INVALID_COMMAND)
                err.SetErrorMessage(msg)

        if self.Parent:
            evt = OutputBufferEvent(edEVT_PROCESS_START,
                                    self.Parent.GetId(),
                                    self.LastCommand)
            wx.PostEvent(self.Parent, evt)

        # Read from stdout while there is output from process
        while not err and True:
            if self.abort:
                self.__KillPid(self.Process.pid)
                self.__DoOneRead()
                more = False
                break
            else:
                more = False
                try:
                    more = self.__DoOneRead()
                except wx.PyDeadObjectError:
                    # Our parent window is dead so kill process and return
                    self.__KillPid(self.Process.pid)
                    return

                if not more:
                    break

        # Notify of error in running the process
        if err is not None:
            if self.Parent:
                wx.PostEvent(self.Parent, err)
            result = -1
        else:
            try:
                result = self.Process.wait()
            except OSError:
                result = -1

        # Notify that process has exited
        # Pack the exit code as the events value
        if self.Parent:
            evt = OutputBufferEvent(edEVT_PROCESS_EXIT, self.Parent.GetId(), result)
            wx.PostEvent(self.Parent, evt)

class ProcessThread(ProcessThreadBase):
    """Run a subprocess in a separate thread. Thread posts events back
    to parent object on main thread for processing in the ui.
    @see: EVT_PROCESS_START, EVT_PROCESS_END, EVT_UPDATE_TEXT

    """
    def __init__(self, parent, command, fname='',
                 args=list(), cwd=None, env=dict(),
                 use_shell=True):
        """Initialize the ProcessThread object
        Example:
          >>> myproc = ProcessThread(myframe, '/usr/local/bin/python',
                                     'hello.py', '--version', '/Users/me/home/')
          >>> myproc.start()

        @param parent: Parent Window/EventHandler to receive the events
                       generated by the process.
        @param command: Command string to execute as a subprocess.
        @keyword fname: Filename or path to file to run command on.
        @keyword args: Argument list or string to pass to use with fname arg.
        @keyword cwd: Directory to execute process from or None to use current
        @keyword env: Environment to run the process in (dictionary) or None to
                      use default.
        @keyword use_shell: Specify whether a shell should be used to launch 
                            program or run directly

        """
        super(ProcessThread, self).__init__(parent)

        if isinstance(args, list):
            args = u' '.join([arg.strip() for arg in args])

        # Attributes
        self._cwd = cwd             # Path at which to run from
        self._cmd = dict(cmd=command, file=fname, args=args)
        self._use_shell = use_shell

        # Make sure the environment is sane it must be all strings
        nenv = dict(env) # make a copy to manipulate
        for k, v in env.iteritems():
            if isinstance(v, types.UnicodeType):
                nenv[k] = v.encode(sys.getfilesystemencoding())
            elif not isinstance(v, basestring):
                nenv.pop(k)
        self._env = nenv

        # Setup
        self.setDaemon(True)

    def DoPopen(self):
        """Open the process
        @return: subprocess.Popen instance

        """
        # using shell, Popen will need a string, else it must be a sequence
        # use shlex for complex command line tokenization/parsing
        command = u' '.join([item.strip() for item in [self._cmd['cmd'],
                                                       self._cmd['file'],
                                                       self._cmd['args']]])
        command = command.strip()
        # TODO: exception handling and notification to main thread
        #       when encoding fails.
        command = command.encode(sys.getfilesystemencoding())
        if not self._use_shell and not subprocess.mswindows:
            # Note: shlex does not support Unicode
            command = shlex.split(command)

        # TODO: if a file path to the exe has any spaces in it on Windows
        #       and use_shell is True then the command will fail. Must force
        #       to False under this condition.
        use_shell = self._use_shell
        # TODO: See about supporting use_shell on Windows it causes lots of
        #       issues with gui apps and killing processes when it is True.
        if use_shell and subprocess.mswindows:
            suinfo = subprocess.STARTUPINFO()
            # Don't set this flag if we are not using the shell on
            # Windows as it will cause any gui app to not show on the
            # display!
            #TODO: move this into common library as it is needed
            #      by most code that uses subprocess
            if hasattr(subprocess, 'STARTF_USESHOWWINDOW'):
                suinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            else:
                try:
                    from win32process import STARTF_USESHOWWINDOW
                    suinfo.dwFlags |= STARTF_USESHOWWINDOW
                except ImportError:
                    # Give up and try hard coded value from Windows.h
                    suinfo.dwFlags |= 0x00000001
        else:
            suinfo = None
        
        proc = subprocess.Popen(command,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                shell=use_shell,
                                cwd=self._cwd,
                                env=self._env,
                                startupinfo=suinfo)
        self.LastCommand = command # Set last run command
        return proc

    def SetArgs(self, args):
        """Set the args to pass to the command
        @param args: list or string of program arguments

        """
        if isinstance(args, list):
            u' '.join(item.strip() for item in args)
        self._cmd['args'] = args.strip()

    def SetCommand(self, cmd):
        """Set the command to execute
        @param cmd: Command string

        """
        self._cmd['cmd'] = cmd

    def SetFilename(self, fname):
        """Set the filename to run the command on
        @param fname: string or Unicode

        """
        self._cmd['file'] = fname

#-----------------------------------------------------------------------------#

class TaskThread(threading.Thread):
    """Run a task in its own thread."""
    def __init__(self, parent, task, *args, **kwargs):
        """Initialize the TaskThread. All *args and **kwargs are passed
        to the task.

        @param parent: Parent Window/EventHandler to receive the events
                       generated by the process.
        @param task: callable should be a generator object and must be iterable

        """
        super(TaskThread, self).__init__()
        assert isinstance(parent, OutputBuffer)

        self._task = TaskObject(parent, task, *args, **kwargs)

    def run(self):
        self._task.DoTask()

    def Cancel(self):
        self._task.Cancel()

class TaskObject(object):
    """Run a task in its own thread."""
    def __init__(self, parent, task, *args, **kwargs):
        """Initialize the TaskObject. All *args and **kwargs are passed
        to the task.

        @param parent: Parent Window/EventHandler to receive the events
                       generated by the process.
        @param task: callable should be a generator object and must be iterable

        """
        super(TaskObject, self).__init__()

        assert isinstance(parent, OutputBuffer)

        # Attributes
        self.cancel = False         # Abort task
        self._parent = parent       # Parent Window/Event Handler
        self.task = task            # Task method to run
        self._args = args
        self._kwargs = kwargs

    def DoTask(self):
        """Start running the task"""
        # Notify that task is beginning
        evt = OutputBufferEvent(edEVT_TASK_START, self._parent.GetId())
        wx.PostEvent(self._parent, evt)
        time.sleep(.5) # Give the event a chance to be processed

        # Run the task and post the results
        for result in self.task(*self._args, **self._kwargs):
            self._parent.AppendUpdate(result)
            if self.cancel:
                break

        # Notify that the task is finished
        evt = OutputBufferEvent(edEVT_TASK_COMPLETE, self._parent.GetId())
        wx.PostEvent(self._parent, evt)

    def Cancel(self):
        """Cancel the running task"""
        self.cancel = True



###############################################################################
# Name: nsis.py                                                               #
# Purpose: Define NSIS syntax for highlighting and other features             #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: nsis.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for Nullsoft Installer Scripts.
@todo: Add User Defined KW

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _nsis.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

# NSIS Functions
NSIS_FUNCT = (0, "!addincludedir !addplugindir MakeNSIS Portions Contributors: "
                 "Abort AddBrandingImage AddSize AutoCloseWindow BGFont "
                 "BrandingText BringToFront Call CallInstDLL Caption ChangeUI "
                 "ClearErrors ComponentText GetDLLVersion GetDLLVersionLocal "
                 "GetFileTime GetFileTimeLocal CopyFiles CRCCheck FileRead "
                 "CreateFont CreateShortCut SetDatablockOptimize DeleteINISec "
                 "DeleteINIStr DeleteRegKey DeleteRegValue Delete DetailPrint "
                 "DirText DirShow DirVar DirVerify GetInstDirError BGGradient"
                 "AllowRootDirInstall CheckBitmap EnableWindow EnumRegKey "
                 "EnumRegValue Exch Exec ExecWait ExecShell ExpandEnvStrings "
                 "FindWindow FindClose FindFirst FindNext File FileBufSize "
                 "FlushINI ReserveFile FileClose FileErrorText FileOpen IntCmp "
                 "FileWrite FileReadByte FileWriteByte FileSeek Function Page "
                 "GetDlgItem GetFullPathName GetTempFileName HideWindow Icon "
                 "IfErrors IfFileExists IfRebootFlag IfSilent InstallDirRegKey "
                 "InstallColors InstallDir InstProgressFlags InstType IntOp "
                 "IntCmpU IntFmt IsWindow Goto LangString LangStringUP Return "
                 "LicenseForceSelection LicenseLangString LicenseText "
                 "LoadLanguageFile LogSet LogText MessageBox Nop Name OutFile "
                 "PageCallbacks PageEx PageExEnd Pop Push Quit ReadINIStr "
                 "ReadRegDWORD ReadRegStr ReadEnvStr Reboot RegDLL Rename "
                 "RMDir Section SectionEnd SectionIn SubSection SectionGroup "
                 "SubSectionEnd SectionGroupEnd SearchPath SectionSetFlags "
                 "SectionGetFlags SectionSetInstTypes SectionGetInstTypes "
                 "SectionGetText SectionSetText SectionGetSize SectionSetSize "
                 "GetCurInstType SetCurInstType InstTypeSetText SetCompress"
                 "SendMessage SetAutoClose SetCtlColors SetBrandingImage  "
                 "SetCompressor SetCompressorDictSize SetCompressionLevel "
                 "SetDateSave SetDetailsView SetDetailsPrint SetErrors "
                 "GetErrorLevel SetFileAttributes SetFont SetOutPath "
                 "SetPluginUnload SetRebootFlag SetShellVarContext SetSilent "
                 "ShowInstDetails ShowUninstDetails ShowWindow SilentInstall "
                 "SilentUnInstall Sleep StrCmp StrCpy StrLen SubCaption "
                 "UninstallExeName UninstallCaption UninstallIcon UninstPage "
                 "UninstallText UninstallSubCaption UnRegDLL WindowIcon "
                 "WriteRegBin WriteRegDWORD WriteRegStr WriteRegExpandStr "
                 "WriteUninstaller XPStyle !packhdr !system !execute !echo "
                 "!include !cd !ifdef !ifndef !endif !define !undef !else "
                 "!warning !error !verbose !macro !macroend !insertmacro "
                 "!ifmacrondef MiscButtonText DetailsButtonText "
                 "InstallButtonText SpaceTexts CompletedText InitPluginsDir "
                 "GetLabelAddress GetCurrentAddress !AddPluginDir LockWindow "
                 "AllowSkipFiles Var VIAddVersionKey VIProductVersion "
                 "ShowUnInstDetails WriteIniStr CreateDirectory FunctionEnd "
                 "IfAbort LicenseData LicenseBkColor InstTypeGetText "
                 "SetErrorLevel SetOverwrite WriteINIStr !AddIncludeDir "
                 "!ifmacrodef UninstallButtonText GetFunctionAddress ")

# NSIS Variables/Constants
NSIS_VAR = (1, "$0 $1 $2 $3 $4 $5 $6 $7 $8 $9 $R0 $R1 $R2 $R3 $R4 $R5 $R6 $R7 "
               "$R8 $R9 $\t $\" $\' $\` $VARNAME $0, $INSTDIR $OUTDIR $CMDLINE "
               "$LANGUAGE $PROGRAMFILES $COMMONFILES $DESKTOP $EXEDIR "
               "${NSISDIR} $WINDIR $SYSDIR $TEMP $STARTMENU $SMPROGRAMS "
               "$SMSTARTUP $QUICKLAUNCH $DOCUMENTS $SENDTO $RECENT $FAVORITES "
               "$MUSIC $PICTURES $VIDEOS $NETHOOD $FONTS $TEMPLATES $APPDATA "
               "$PRINTHOOD $INTERNET_CACHE $COOKIES $HISTORY $PROFILE "
               "$ADMINTOOLS $RESOURCES $RESOURCES_LOCALIZED $CDBURN_AREA "
               "$HWNDPARENT $PLUGINSDIR $$ $\r $\n")

# NSIS Lables (Attributes)
NSIS_LBL = (2, "ARCHIVE FILE_ATTRIBUTE_ARCHIVE FILE_ATTRIBUTE_HIDDEN LEFT "
               "FILE_ATTRIBUTE_NORMAL FILE_ATTRIBUTE_OFFLINE lastused HKCR "
               "FILE_ATTRIBUTE_SYSTEM FILE_ATTRIBUTE_TEMPORARY HIDDEN HKCC "
               "HKCU HKDD HKEY_CLASSES_ROOT HKEY_CURRENT_CONFIG IDYES SYSTEM "
               "HKEY_DYN_DATA HKEY_LOCAL_MACHINE MB_ICONQUESTION OFFLINE "
               "HKLM HKPD HKU IDABORT IDCANCEL IDIGNORE IDNO IDOK IDRETRY "
               "MB_ABORTRETRYIGNORE MB_DEFBUTTON1 MB_DEFBUTTON2 MB_DEFBUTTON3 "
               "MB_DEFBUTTON4 MB_ICONEXCLAMATION MB_ICONINFORMATION normal off "
               "MB_ICONSTOP MB_OK MB_OKCANCEL MB_RETRYCANCEL MB_RIGHT listonly "
               "MB_SETFOREGROUND MB_TOPMOST MB_YESNO MB_YESNOCANCEL NORMAL "
               "READONLY SW_SHOWMAXIMIZED SW_SHOWMINIMIZED SW_SHOWNORMAL "
               "TEMPORARY auto colored false force hide ifnewer nevershow "
               "on show silent silentlog smooth true try lzma zlib bzip2 none "
               "textonly both top left bottom right license components "
               "instfiles uninstConfirm custom all leave current ifdiff "
               "RIGHT CENTER dlg_id ALT CONTROL EXT SHIFT open print manual "
               "alwaysoff FILE_ATTRIBUTE_READONLY HKEY_CURRENT_USER directory "
               "HKEY_PERFORMANCE_DATA HKEY_USERS ")

# NSIS User Defined (Not sure need help)
NSIS_DEF = (3, "")

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (stc.STC_NSIS_DEFAULT, 'default_style'),
                 (stc.STC_NSIS_COMMENT, 'comment_style'),
                 (stc.STC_NSIS_FUNCTION, 'funct_style'),
                 (stc.STC_NSIS_FUNCTIONDEF, 'keyword_style'),
                 (stc.STC_NSIS_IFDEFINEDEF, 'pre_style'),
                 (stc.STC_NSIS_LABEL, 'class_style'),
                 (stc.STC_NSIS_MACRODEF, 'pre_style'),
                 (stc.STC_NSIS_NUMBER, 'number_style'),
                 (stc.STC_NSIS_SECTIONDEF, 'keyword_style'),
                 (stc.STC_NSIS_STRINGDQ, 'string_style'),
                 (stc.STC_NSIS_STRINGLQ, 'string_style'),
                 (stc.STC_NSIS_STRINGRQ, 'string_style'),
                 (stc.STC_NSIS_STRINGVAR, 'string_style'),
                 (stc.STC_NSIS_SUBSECTIONDEF, 'keyword_style'),
                 (stc.STC_NSIS_USERDEFINED, 'pre_style'),
                 (stc.STC_NSIS_VARIABLE, 'scalar_style') ]

#---- Extra Properties ----#
FOLD = ("fold", "1")

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for NSIS""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_NSIS)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [NSIS_FUNCT, NSIS_VAR, NSIS_LBL]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u';']



###############################################################################
# Name: batch.py                                                              #
# Purpose: Define DOS Batch syntax for highlighting and other features        #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: batch.py
AUTHOR: Cody Precord
@summary: Lexer configuration file for dos/windows batch scripts.
@todo: incorportate winbat keywords

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _batch.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

DOSBAT_KEYWORDS = (0, "append arp assoc at atmadm attrib bootcfg break cacls "
                      "call cd chcp chdir chkdsk chkntfs cls cmd color command "
                      "comp compact control convert copy date debug defrag del "
                      "delete dir diskcomp diskcopy doskey echo edit edlin "
                      "else endlocal EQU erase errorlevel exists exit expand "
                      "fc find findstr for format ftp ftype GEQ goto graftabl "
                      "GTR help if ipconfig keyb label LEQ loadfix logoff LSS "
                      "md mkdir mode more move nbtstat NEQ net netsh netstat "
                      "nlsfunc not nslookup path pathping pause ping popd "
                      "print prompt pushd rd rem ren rename replace restore "
                      "rmdir route runas set setlocal setver share shift "
                      "shutdown sort start subst time title telnet tracetr "
                      "tree type ver verify vol xcopy ")

# WinBatch Keywords
WINBAT_KEYWORDS = (0, "if then else endif break end return exit next while for "
                      "gosub goto switch select to case endselect endwhile "
                      "endswitch aboveicons acc_attrib acc_chng_nt acc_control "
                      "acc_create acc_delete acc_full_95 acc_full_nt acc_list "
                      "acc_pfull_nt acc_pmang_nt acc_print_nt acc_read "
                      "acc_read_95 acc_read_nt acc_write amc arrange ascending "
                      "attr_a attr_a attr_ci attr_ci attr_dc attr_dc attr_di "
                      "attr_di attr_dm attr_dm attr_h attr_h attr_ic attr_ic "
                      "attr_p attr_p attr_ri attr_ri attr_ro attr_ro attr_sh "
                      "attr_sh attr_sy attr_sy attr_t attr_t attr_x attr_x "
                      "avogadro backscan boltzmann cancel capslock check "
                      "columnscommonformat cr crlf ctrl default default "
                      "deg2rad descending disable drive electric enable eulers "
                      "false faraday float8 fwdscan gftsec globalgroup gmtsec "
                      "goldenratio gravitation hidden icon lbutton lclick "
                      "ldblclick lf lightmps lightmtps localgroup magfield "
                      "major mbokcancel mbutton mbyesno mclick mdblclick minor "
                      "msformat multiple ncsaformat no none none noresize "
                      "normal notify nowait numlock off on open parsec "
                      "parseonly pi planckergs planckjoules printer rad2deg "
                      "rbutton rclick rdblclick regclasses regcurrent "
                      "regmachine regroot regusers rows save scrolllock server "
                      "shift single sorted stack string tab tile true uncheck "
                      "unsorted wait wholesection word1 word2 word4 yes zoomed "
                      "about abs acos addextender appexist appwaitclose asin "
                      "askfilename askfiletext askitemlist askline askpassword "
                      "askyesno atan average beep binaryalloc binarycopy "
                      "binaryeodget binaryeodset binaryfree binaryhashrec "
                      "binaryincr binaryincr2 binaryincr4 binaryincrflt "
                      "binaryindex binaryindexnc binaryoletype binarypeek "
                      "binarypeek2 binarypeek4 binarypeekflt binarypeekstr "
                      "binarypoke binarypoke2 binarypoke4 binarypokeflt "
                      "binarypokestr binaryread binarysort binarystrcnt "
                      "binarywrite boxbuttondraw boxbuttonkill boxbuttonstat "
                      "boxbuttonwait boxcaption boxcolor boxdataclear "
                      "boxdatatag boxdestroy boxdrawcircle boxdrawline "
                      "boxdrawrect boxdrawtext boxesup boxmapmode boxnew "
                      "boxopen boxpen boxshut boxtext boxtextcolor boxtextfont "
                      "boxtitle boxupdates break buttonnames by call callext "
                      "ceiling char2num clipappend clipget clipput continue "
                      "cos cosh datetime ddeexecute ddeinitiate ddepoke "
                      "dderequest ddeterminate ddetimeout debug debugdata "
                      "decimals delay dialog dialogbox dirattrget dirattrset "
                      "dirchange direxist")

#---- Language Styling Specs ----#
SYNTAX_ITEMS = [ (stc.STC_BAT_DEFAULT,  "default_style"),
                 (stc.STC_BAT_COMMAND,  "class_style"),
                 (stc.STC_BAT_COMMENT,  "comment_style"),
                 (stc.STC_BAT_HIDE,     "string_style"),
                 (stc.STC_BAT_IDENTIFIER, "scalar_style"),
                 (stc.STC_BAT_LABEL,    "class_style"),
                 (stc.STC_BAT_OPERATOR, "operator_style"),
                 (stc.STC_BAT_WORD,     "keyword_style") ]

#---- Extra Properties ----#
FOLD = ("fold", "1")

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Batch files""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_BATCH)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [DOSBAT_KEYWORDS]

    def GetSyntaxSpec(self):
        """Syntax Specifications"""
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set"""
        return [FOLD]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'::']



###############################################################################
# Name: python.py                                                             #
# Purpose: Define Python syntax for highlighting and other features           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: python.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for Python.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _python.py 69016 2011-09-06 20:00:06Z CJP $"
__revision__ = "$Revision: 69016 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc
import keyword

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

# Indenter keywords
INDENT_KW = (u"def", u"if", u"elif", u"else", u"for", u"while",
             u"class", u"try", u"except", u"finally", u"with")
UNINDENT_KW = (u"return", u"raise", u"break", u"continue",
               u"pass", u"exit", u"quit")

# Python Keywords
KEYWORDS = keyword.kwlist
KEYWORDS.extend(['True', 'False', 'None', 'self'])
PY_KW = (0, u" ".join(KEYWORDS))

# Highlighted builtins
try:
    import __builtin__
    BUILTINS = dir(__builtin__)
except:
    BUILTINS = list()
#BUILTINS.append('self')
BUILTINS = list(set(BUILTINS))

PY_BIN = (1, u" ".join(sorted(BUILTINS)))

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (stc.STC_P_DEFAULT, 'default_style'),
                 (stc.STC_P_CHARACTER, 'char_style'),
                 (stc.STC_P_CLASSNAME, 'class_style'),
                 (stc.STC_P_COMMENTBLOCK, 'comment_style'),
                 (stc.STC_P_COMMENTLINE, 'comment_style'),
                 (stc.STC_P_DECORATOR, 'decor_style'),
                 (stc.STC_P_DEFNAME, 'keyword3_style'),
                 (stc.STC_P_IDENTIFIER, 'default_style'),
                 (stc.STC_P_NUMBER, 'number_style'),
                 (stc.STC_P_OPERATOR, 'operator_style'),
                 (stc.STC_P_STRING, 'string_style'),
                 (stc.STC_P_STRINGEOL, 'stringeol_style'),
                 (stc.STC_P_TRIPLE, 'string_style'),
                 (stc.STC_P_TRIPLEDOUBLE, 'string_style'),
                 (stc.STC_P_WORD, 'keyword_style'),
                 (stc.STC_P_WORD2, 'userkw_style')]

#---- Extra Properties ----#
FOLD = ("fold", "1")
FOLD_QUOTES = ("fold.quotes.python", "1")
FOLD_COMMENTS = ("fold.comment.python", "1")
TIMMY = ("tab.timmy.whinge.level", "1") # Mark Inconsistent indentation

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Python""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_PYTHON)
        self.RegisterFeature(synglob.FEATURE_AUTOINDENT, AutoIndenter)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [PY_KW, PY_BIN]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD, TIMMY, FOLD_QUOTES, FOLD_COMMENTS]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'#']

#-----------------------------------------------------------------------------#

def AutoIndenter(estc, pos, ichar):
    """Auto indent python code.
    @param estc: EditraStyledTextCtrl
    @param pos: current carat position
    @param ichar: Indentation character

    """
    line = estc.GetCurrentLine()
    spos = estc.PositionFromLine(line)
    text = estc.GetTextRange(spos, pos)
    eolch = estc.GetEOLChar()
    inspace = text.isspace()

    # Cursor is in the indent area somewhere
    if inspace:
        estc.AddText(eolch + text.replace(eolch, u""))
        return

    # Check if the cursor is in column 0 and just return newline.
    if not len(text):
        estc.AddText(eolch)
        return

    # In case of open bracket: Indent next to open bracket
    def BackTrack(tmp_text, tline):
        bcount = [ tmp_text.count(brac) for brac in u")}]({[" ]
        bRecurse = False
        for idx, val in enumerate(bcount[:3]):
            if val > bcount[idx+3]:
                bRecurse = True
                break
        if bRecurse:
            tline = tline - 1
            if tline < 0:
                return tmp_text
            spos = estc.PositionFromLine(tline)
            tmp_text = estc.GetTextRange(spos, pos)
            BackTrack(tmp_text, tline)
        return tmp_text
    text = BackTrack(text, line)
    pos = PosOpenBracket(text)
    if pos > -1:
        rval = eolch + (pos + 1) * u" "
        estc.AddText(rval)
        return

    indent = estc.GetLineIndentation(line)
    if ichar == u"\t":
        tabw = estc.GetTabWidth()
    else:
        tabw = estc.GetIndent()

    i_space = indent / tabw
    end_spaces = ((indent - (tabw * i_space)) * u" ")

    tokens = filter(None, text.strip().split())
    if tokens and not inspace:
        if tokens[-1].endswith(u":"):
            if tokens[0].rstrip(u":") in INDENT_KW:
                i_space += 1
        elif tokens[-1].endswith(u"\\"):
            i_space += 1
        elif len(tokens[-1]) and tokens[-1][-1] in u"}])":
            ptok = tokens[-1][-1]
            paren_pos = pos - (len(text) - text.rfind(ptok))
            oparen, cparen = estc.GetBracePair(paren_pos)
            if cparen >= 0: # Found matching bracket
                line = estc.LineFromPosition(cparen)
                indent = estc.GetLineIndentation(line)
                i_space = indent / tabw
                end_spaces = ((indent - (tabw * i_space)) * u" ")
        elif tokens[0] in UNINDENT_KW:
            i_space = max(i_space - 1, 0)

    rval = eolch + (ichar * i_space) + end_spaces
    if inspace and ichar != u"\t":
        rpos = indent - (pos - spos)
        if rpos < len(rval) and rpos > 0:
            rval = rval[:-rpos]
        elif rpos >= len(rval):
            rval = eolch

    # Put text in the buffer
    estc.AddText(rval)

#---- End Required Module Functions ----#

#---- Syntax Modules Internal Functions ----#
def KeywordString():
    """Returns the specified Keyword String
    @note: not used by most modules

    """
    return PY_KW[1]

def PosOpenBracket(text):
    """Returns the position of the right most open bracket in text.
    Brackets inside strings are ignored. In case of no open bracket
    the returned value is -1
    @param text: The line preceding the new line to be indented.
    @return res: The position of right most open bracket.
    @note: Used by AutoIndenter

    """
    # Store positions of '(', '[','{', ')', ']', '}'
    brackets = [[], [], [], [], [], []]
    quotes = u"'" + u'"'
    in_string = False
    for pos, char in enumerate(text):
        if in_string:
            in_string = not char == quote
        else:
            if char == u'#':
                break
            typ = u'([{)]}'.find(char)
            if typ > -1:
                brackets[typ].append(pos)
            else:
                typ = quotes.find(char)
                if typ > -1:
                    in_string = True
                    quote = quotes[typ]
    res = -1
    for typ in range(3):
        opn, cls = brackets[typ], brackets[typ+3]
        nopn, ncls = len(opn), len(cls)
        if nopn > ncls:
            res = max(res, opn[nopn - ncls - 1])
    return res

#---- End Syntax Modules Internal Functions ----#



###############################################################################
# Name: ed_session.py                                                         #
# Purpose: Object to help manage editor sessions                              #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2011 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Editra session file manager.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: ed_session.py 70097 2011-12-22 21:36:43Z CJP $"
__revision__ = "$Revision: 70097 $"

#-----------------------------------------------------------------------------#
# Imports
import wx
import os
import cPickle

# Editra Imports
import util
import ed_cmdbar
import ed_glob
import profiler
import ed_msg
import ebmlib

#-----------------------------------------------------------------------------#
# Globals
_ = wx.GetTranslation

#-----------------------------------------------------------------------------#

class SessionManager(object):
    """Simple editing session manager helper class"""
    def __init__(self, savedir):
        """@param savedir: directory to load/save session files at"""
        super(SessionManager, self).__init__()

        # Attributes
        self.__default = '__default' # default session file name
        self._sessiondir = savedir
        self._sessionext = '.session'

    #---- Properties ----#
    DefaultSession = property(lambda self: self.__default)

    SessionDir = property(lambda self: self._sessiondir,
                          lambda self, dpath: setattr(self, '_sessiondir', dpath))

    def _SetExtension(self, ext):
        assert ext.startswith('.')
        self._sessionext = ext
    SessionExtension = property(lambda self: self._sessionext,
                                lambda self, ext: self._SetExtension(ext))
    Sessions = property(lambda self: self.GetSavedSessions())

    #---- Implementation ----#

    def DeleteSession(self, name):
        """Delete the specified session name
        @param name: session name
        @return: bool

        """
        rval = True
        session = self.PathFromSessionName(name)
        if os.path.exists(session):
            try:
                os.remove(session)
            except OSError:
                rval = False
        return rval

    def GetSavedSessions(self):
        """Get the list of available saved sessions by display name
        @return: list of strings

        """
        sessions = list()
        defaultSession = None
        for session in os.listdir(self.SessionDir):
            if session.endswith(self.SessionExtension):
                path = os.path.join(self.SessionDir, session)
                sName = self.SessionNameFromPath(path)
                if session.startswith(self.__default):
                    defaultSession = sName
                else:
                    sessions.append(sName)
        sessions.sort()
        if defaultSession:
            sessions.insert(0, defaultSession)
        return sessions

    def LoadSession(self, name):
        """Load a named session
        @param name: session name
        @return: list of paths

        """
        session = self.PathFromSessionName(name)
        assert os.path.exists(session)
        flist = list()
        with open(session, 'rb') as f_handle:
            # Load and validate file
            try:
                flist = cPickle.load(f_handle)
                # TODO: Extend in future to support loading sessions
                #       for multiple windows.
                flist = flist.get('win1', list())
                for item in flist:
                    if type(item) not in (unicode, str):
                        raise TypeError("Invalid item in unpickled sequence")
            except (cPickle.UnpicklingError, TypeError, EOFError), e:
                util.Log("[ed_session][err] %s" % e)
                raise e # Re throw
        return flist

    def SaveSession(self, name, paths):
        """Save the given list of files as a session with the given name
        @param name: session name
        @param paths: list of file paths
        @return: bool

        """
        session = self.PathFromSessionName(name)
        bOk = False
        with open(session, 'wb') as f_handle:
            try:
                # TODO multi window support
                sdata = dict(win1=paths)
                cPickle.dump(sdata, f_handle)
                bOk = True
            except Exception, msg:
                util.Log("[ed_session][err] Failed to SaveSessionFile: %s" % msg)

        return bOk

    def PathFromSessionName(self, session):
        """Get the full path to store a session file
        @param session: string base name (no extension)

        """
        name = session + self.SessionExtension
        path = os.path.join(self.SessionDir, name)
        return path

    def SessionNameFromPath(self, path):
        """Get the sessions display name from its path"""
        assert path.endswith(self.SessionExtension)
        name = os.path.basename(path)
        name = name.rsplit('.', 1)[0]
        return name

#-----------------------------------------------------------------------------#

class EdSessionMgr(SessionManager):
    """Editra specific session manager implementation"""
    __metaclass__ = ebmlib.Singleton
    def __init__(self):
        super(EdSessionMgr, self).__init__(ed_glob.CONFIG['SESSION_DIR'])

#-----------------------------------------------------------------------------#

class EdSessionBar(ed_cmdbar.CommandBarBase):
    """Command bar for managing editing sessions"""
    class meta:
        id = ed_glob.ID_SESSION_BAR

    def __init__(self, parent):
        super(EdSessionBar, self).__init__(parent)

        # Attributes
        self._sch = wx.Choice(self)

        # Setup
        self.__DoLayout()
        self._saveb = self.AddPlateButton(_("Save"), ed_glob.ID_SAVE)
        self._saveb.Name = "SaveButton"
        self._saveasb = self.AddPlateButton(_("Save As"), ed_glob.ID_SAVEAS)
        self._saveasb.Name = "SaveAsButton"
        self._delb = self.AddPlateButton(_("Delete"), ed_glob.ID_DELETE)
        self._delb.Name = "DeleteButton"
        self.UpdateSessionList()

        # Event Handlers
        self._saveb.Bind(wx.EVT_BUTTON, self.OnSaveSession)
        self._saveasb.Bind(wx.EVT_BUTTON, self.OnSaveSession)
        self._delb.Bind(wx.EVT_BUTTON, self.OnDeleteSession)
        self._sch.Bind(wx.EVT_CHOICE, self.OnChangeSession)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUI, self._delb)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy)

        # Message Handlers
        ed_msg.Subscribe(self.OnConfigMsg,
                         ed_msg.EDMSG_PROFILE_CHANGE + ('LAST_SESSION',))

    def __DoLayout(self):
        self.AddControl(wx.StaticText(self, label=_("Session:")), wx.ALIGN_LEFT)
        self.AddControl(self._sch, wx.ALIGN_LEFT)

    def OnDestroy(self, evt):
        """Cleanup message handlers"""
        if self:
            ed_msg.Unsubscribe(self.OnConfigMsg)
        evt.Skip()

    def GetSelectedSession(self):
        """Get the currently selected session
        @return: string (internal session name)

        """
        if self._sch.Selection == 0:
            return EdSessionMgr().DefaultSession
        else:
            return self._sch.StringSelection

    def OnChangeSession(self, evt):
        """Current session changed in choice control"""
        util.Log(u"[ed_session][info] OnChangeSession: %s" % self._sch.StringSelection)
        ed_msg.PostMessage(ed_msg.EDMSG_SESSION_DO_LOAD, 
                           self.GetSelectedSession(),
                           self.TopLevelParent.Id)

    def OnConfigMsg(self, msg):
        """Configuration update callback"""
        if self:
            self.UpdateSessionList()

    def OnSaveSession(self, evt):
        """Save the current session"""
        ofiles = list()
        util.Log("[ed_session][info] OnSaveSession")
        if evt.EventObject is self._saveb:
            ed_msg.PostMessage(ed_msg.EDMSG_FILE_GET_OPENED, ofiles,
                               self.TopLevelParent.Id)
            util.Log("[ed_session][info] OnSaveSession: %d files" % len(ofiles))
            EdSessionMgr().SaveSession(self.GetSelectedSession(), ofiles)
        elif evt.EventObject is self._saveasb:
            ed_msg.PostMessage(ed_msg.EDMSG_SESSION_DO_SAVE,
                               context=self.TopLevelParent.Id)
            # Bar will be updated by config change if the save succeeds
        else:
            evt.Skip()

    def OnDeleteSession(self, evt):
        """Delete the current session"""
        if evt.EventObject is self._delb:
            ses = self.GetSelectedSession()
            if ses != EdSessionMgr().DefaultSession:
                EdSessionMgr().DeleteSession(ses)
                # Switch back to default session
                profiler.Profile_Set('LAST_SESSION',
                                     EdSessionMgr().DefaultSession)
        else:
            evt.Skip()

    def OnBarShown(self):
        """Update the session list"""
        self.UpdateSessionList()

    def OnUpdateUI(self, evt):
        """Handle UpdateUI events"""
        if evt.EventObject is self._delb:
            evt.Enable(self._sch.Selection > 0)
        else:
            evt.Skip()

    def UpdateSessionList(self):
        """Update the session list"""
        sessions = EdSessionMgr().Sessions
        if len(sessions):
            sessions[0] = _("Default")
            self._sch.Items = sessions
            self.UpdateSelectedSession()
            self.Layout()
            self.Refresh()

    def UpdateSelectedSession(self):
        """Select the currently configured session"""
        ses = profiler.Profile_Get('LAST_SESSION')
        if ses:
            if ses == EdSessionMgr().DefaultSession:
                self._sch.SetSelection(0)
            elif ses in self._sch.Items:
                self._sch.SetStringSelection(ses)



#!/usr/bin/env python
###############################################################################
# Name: gen_lexer_doc.py                                                      #
# Purpose: Generate html for lexer docuemntation                              #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Lexer Documentation Generator

Generates html documentation for the STC lexers and style flags.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: gen_stc_doc.py 63082 2010-01-05 21:17:26Z CJP $"
__revision__ = "$Revision: 63082 $"

#-----------------------------------------------------------------------------#
# Imports
import sys
import wx
import wx.stc as stc

#-----------------------------------------------------------------------------#

# Special case lexer to syntax spec mappings
LEXMAP = { 'BASH'     : ['SH',],
           'BATCH'    : ['BAT',],
           'CPP'      : ['C',],
           'F77'      : ['F',],
           'FLAGSHIP' : ['FS',],
           'FORTRAN'  : ['F',],
           'GUI4CLI'  : ['GC',],
           'HASKELL'  : ['HA',],
           'HTML'     : ['H', 'HJ', 'HJA', 'HBA'],
           'INNOSETUP': ['INNO',],
           'LATEX'    : ['TEX',],
           'MAKEFILE' : ['MAKE',],
           'OCTAVE'   : ['MATLAB',],
           'PASCAL'   : ['C',],
           'PERL'     : ['PL',],
           'PROPERTIES' : ['PROPS',],
           'PYTHON'   : ['P',],
           'RUBY'     : ['RB',],
           'SMALLTALK': ['ST',],
           'VB'       : ['B',],
           'VBSCRIPT' : ['B',],
           'VERILOG'  : ['V',],
           'XML'      : ['H',],
          }

#-----------------------------------------------------------------------------#
# HTML objects

class Title(object):
    """Represents a title tag"""
    def __init__(self, title, mode):
        assert isinstance(title, basestring)
        object.__init__(self)

        # Attributes
        self.mode = mode
        self._title = title

    def __str__(self):
        if self.mode == 'html':
            val = "<h3 id=\"%s\">%s:</h3>\n" % (self._title, self._title)
        elif self.mode == 'moin':
            val = "<<Anchor(%s)>>\n== %s ==\n" % (self._title, self._title)
        return val

class LexerLabel(object):
    """Represents a lexer id label"""
    def __init__(self, lexer, mode):
        assert isinstance(lexer, basestring)
        object.__init__(self)

        # Attributes
        self.mode = mode
        self._lexer = lexer

    def __str__(self):
        if self.mode == 'html':
            val = "<h4>Lexer Id: <em>%s</em></h4>\n" % self._lexer
        else:
            val = "==== Lexer Id: %s ====\n" % self._lexer
        return val

class UnorderedList(object):
    """Represents a <ul> </ul> list"""
    def __init__(self, items, mode):
        assert isinstance(items, list)
        object.__init__(self)

        # Attributes
        self.mode = mode
        self._items = items

    def __str__(self):
        if self.mode == 'html':
            val = "<ul>\n%s\n</ul>" % self.FormatItems()
        elif self.mode == 'moin':
            val = "%s" % self.FormatItems()
        return val

    def FormatItems(self):
        """Format the item list
        @return: string

        """
        if self.mode == 'html':
            tmpl = "<li>%s</li>\n"
        else:
            tmpl = " * %s\n"
        rval = ""
        for item in self._items:
            rval += tmpl % item
        return rval

class Index(object):
    """Represents the link index"""
    def __init__(self, langs, mode):
        """@param langs: list of languages"""
        object.__init__(self)

        # Attrbutes
        self.mode = mode
        self._langs = langs

    def __str__(self):
        """Create the html table"""
        if self.mode == 'html':
            tmpl = "<table cellpadding=\"3px\" cellspacing=\"0\" border=\"0\" width=\"100%%\">"
            tmpl += "<tr valign=\"top\">"
            tmpl += "<td>%s</td><td>%s</td><td>%s</td>\n"
            tmpl += "</tr>\n</table>\n<hr/>\n"
        elif self.mode == 'moin':
            return "<<TableOfContents(2)>>\n"
        div = len(self._langs) / 3
        items1 = self.makeLinks(self._langs[:div])
        items2 = self.makeLinks(self._langs[div:div+div])
        items3 = self.makeLinks(self._langs[-1*div:])
        return tmpl % (UnorderedList(items1, mode),
                       UnorderedList(items2, mode),
                       UnorderedList(items3, mode))

    def makeLinks(self, items):
        """Make a list of anchor links
        @param items: list of strings
        @return: list of string

        """
        rval = list()
        if self.mode == 'html':
            rval = ["<a href=\"#%s\">%s</a>" % (item, item) for item in items]
        elif self.mode == 'moin':
            rval = ["[[#%s|%s]]" % (item, item) for item in items]
        return rval

class Intro(object):
    def __init__(self, mode):
        object.__init__(self)

        # Attributes
        self.mode = mode

    def __str__(self):
        rval = ''
        if self.mode == 'html':
            rval = "<h1>StyledTextCtrl Quick Reference</h1>"
            rval += "\n<hr/>\n"
        elif self.mode == 'moin':
            rval = "= StyledTextCtrl Quick Reference =\n----\n"
        return rval

class SubSection(object):
    """Represents a subsection for a given language type"""
    def __init__(self, lang, lex, mode):
        assert isinstance(lang, basestring)
        object.__init__(self)

        # Attributes
        self.mode = mode
        self.langid = lang.upper()
        self.title = Title(lang.title(), mode)
        self.lexer = LexerLabel(lex, mode)

    def __str__(self):
        if self.mode == 'html':
            tmpl = "%s%s%s\n<hr/>"
        elif self.mode == 'moin':
            tmpl = "%s%s%s\n----\n"
        styles = self.GetStyleList()
        return tmpl % (self.title, self.lexer, styles)

    def GetStyleList(self):
        """Get the list of styles available for this subsections language
        @return: UnorderedList

        """
        langids = LEXMAP.get(self.langid, [self.langid,])
        slist = list()
        for langid in langids:
            sid = "STC_%s_" % langid
            for item in dir(stc):
                if item.startswith(sid):
                    slist.append(item)
        slist.sort()
        return UnorderedList(slist, self.mode)

#-----------------------------------------------------------------------------#

def collectLexers():
    """Get all the lexer identifiers
    @return: list of strings

    """
    rlist = list()
    for item in dir(stc):
        if item.startswith('STC_LEX_'):
            rlist.append(item)
    rlist.sort()
    return rlist

def deriveLanguageNames(lexlist):
    """Derive the language names from the lexer identifier list
    @return: list of strings

    """
    rlist = list()
    for item in lexlist:
        lang = item.rsplit('_', 1)[-1]
        rlist.append(lang)
    return rlist

#-----------------------------------------------------------------------------#

def generateOutput(mode):
    """Generates the html output
    @return: list of html objects

    """
    lexers = collectLexers()
    langs = deriveLanguageNames(lexers)
    lang2lex = zip(langs, lexers)
    output = [Intro(mode), Index([lang.title() for lang in langs], mode),]
    for lang, lex in lang2lex:
        sub = SubSection(lang, lex, mode)
        output.append(sub)

    f = open('stcdoc.%s' % mode, 'wb')
    for obj in output:
        f.write(str(obj))
    f.close()

#-----------------------------------------------------------------------------#

if __name__ == '__main__':
    mode = 'html' # default to html
    if len(sys.argv) > 1:
        if sys.argv[1] == 'moin':
            mode = 'moin'
    generateOutput(mode)



###############################################################################
# Name: testSynGlob.py                                                        #
# Purpose: Unit tests for ed_basestc                                          #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2012 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""Unittest cases for testing the STC base class"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id:  $"
__revision__ = "$Revision:  $"

#-----------------------------------------------------------------------------#
# Imports
import os
import unittest
import common

# Module to test
import ed_glob
import ed_basestc

#-----------------------------------------------------------------------------#
# Test Class

class EdBaseStcTest(unittest.TestCase):
    def setUp(self):
        # Setup config
        ed_glob.CONFIG['STYLES_DIR'] = common.GetStylesDir()

        # Create control(s)
        self.frame = common.TestFrame(None)
        self.stc = ed_basestc.EditraBaseStc(self.frame)

    def tearDown(self):
        self.frame.Destroy()

    #---- Test Cases ----#

    def testAddLine(self):
        """Test the AddLine method"""
        ccount = self.stc.LineCount
        cpos = self.stc.CurrentLine
        self.stc.AddLine() # add after
        self.assertTrue(ccount == (self.stc.LineCount - 1)) # should be 1 new line
        self.assertTrue(cpos == (self.stc.CurrentLine - 1),
                        repr((cpos, self.stc.CurrentLine))) # cursor position should move to new line
        ccount = self.stc.LineCount
        cpos = self.stc.CurrentLine
        # Test add line above current one
        self.stc.AddLine(before=True)
        self.assertTrue(ccount == (self.stc.LineCount - 1)) # should be 1 new line
        self.assertTrue(cpos == self.stc.CurrentLine, 
                        repr((cpos, self.stc.CurrentLine))) # cursor position should be same line
        # TODO: add tests for indent

    def testGetEOLChar(self):
        """Test that correct eol character is returned"""
        fresh_stc = ed_basestc.EditraBaseStc(self.frame)
        eolchr = fresh_stc.GetEOLChar()
        fresh_stc.Destroy()
        self.assertEquals(eolchr, os.linesep)

    def testLoadFile(self):
        """Test loading a file into the control"""
        path = common.GetDataFilePath("syntax.xml")
        self.assertTrue(self.stc.LoadFile(path))
        self.assertFalse(self.stc.GetReadOnly()) # should be good text



###############################################################################
# Name: perl.py                                                               #
# Purpose: Define Perl syntax for highlighting and other features             #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: perl.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for Perl.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _perl.py 66108 2010-11-10 21:04:54Z CJP $"
__revision__ = "$Revision: 66108 $"

#-----------------------------------------------------------------------------#
# Imports
import wx
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

# Perl Keywords
PERL_KW = (0, "if elseif unless else switch eq ne gt lt ge le cmp not and or "
              "xor while for foreach do until continue defined undef and or "
              "not bless ref BEGIN END my local our goto return last next redo "
              "chomp chop chr crypt index lc lcfirst length org pack reverse "
              "rindex sprintf substr uc ucfirst pos quotemet split study abs "
              "atan2 cos exp hex int log oct rand sin sqrt srand spice unshift "
              "shift push pop split join reverse grep map sort unpack each "
              "exists keys values tie tied untie carp confess croak dbmclose "
              "dbmopen die syscall binmode close closedir eof fileno getc "
              "lstat print printf readdir readline readpipe rewinddir select "
              "stat tell telldir write fcntl flock ioctl open opendir read "
              "seek seekdir sysopen sysread sysseek syswrite truncate pack vec "
              "chdir chmod chown chroot glob link mkdir readlink rename rmdir "
              "symlink umask ulink utime caller dump eval exit wanarray "
              "import alarm exec fork getpgrp getppid getpriority kill pipe "
              "setpgrp setpriority sleep system times wait waitpid accept "
              "bind connect getpeername getsockname getsockopt listen recv "
              "send setsockopt shutdown socket socketpair msgctl msgget msgrcv "
              "msgsnd semctl semget semop shmctl shmget shmread shmwrite "
              "endhostent endnetent endprooent endservent gethostbyaddr "
              "gethostbyname gethostent getnetbyaddr getnetbyname getnetent "
              "getprotobyname getprotobynumber getprotoent getervbyname time "
              "getservbyport getservent sethostent setnetent setprotoent "
              "setservent getpwuid getpwnam getpwent setpwent endpwent "
              "getgrgid getlogin getgrnam setgrent endgrent gtime localtime "
              "times warn formline reset scalar delete prototype lock new "
              "NULL __FILE__ __LINE__ __PACKAGE__ __DATA__ __END__ AUTOLOAD "
              "BEGIN CORE DESTROY END EQ GE GT INIT LE LT NE CHECK use sub "
              "elsif require getgrent ")

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (stc.STC_PL_DEFAULT, 'default_style'),
                 (stc.STC_PL_ARRAY, 'array_style'),
                 (stc.STC_PL_BACKTICKS, 'btick_style'),
                 (stc.STC_PL_CHARACTER, 'char_style'),
                 (stc.STC_PL_COMMENTLINE, 'comment_style'),
                 (stc.STC_PL_DATASECTION, 'default_style'), # STYLE ME
                 (stc.STC_PL_ERROR, 'error_style'),
                 (stc.STC_PL_HASH, 'global_style'),
                 (stc.STC_PL_HERE_DELIM, 'here_style'),
                 (stc.STC_PL_HERE_Q, 'here_style'),
                 (stc.STC_PL_HERE_QQ, 'here_style'),
                 (stc.STC_PL_HERE_QX, 'here_style'),
                 (stc.STC_PL_IDENTIFIER, 'default_style'),
                 (stc.STC_PL_LONGQUOTE, 'default_style'), # STYLE ME
                 (stc.STC_PL_NUMBER, 'number_style'),
                 (stc.STC_PL_OPERATOR, 'operator_style'),
                 (stc.STC_PL_POD, 'comment_style'),
                 (stc.STC_PL_PREPROCESSOR,  'pre_style' ),
                 (stc.STC_PL_PUNCTUATION, 'default_style'), # STYLE ME
                 (stc.STC_PL_REGEX, 'regex_style'),
                 (stc.STC_PL_REGSUBST, 'regex_style'),
                 (stc.STC_PL_SCALAR, 'scalar_style'),
                 (stc.STC_PL_STRING, 'string_style'),
                 (stc.STC_PL_STRING_Q, 'string_style'),
                 (stc.STC_PL_STRING_QQ, 'string_style'),
                 (stc.STC_PL_STRING_QR, 'string_style'),
                 (stc.STC_PL_STRING_QW, 'string_style'),
                 (stc.STC_PL_STRING_QX, 'string_style'),
                 (stc.STC_PL_SYMBOLTABLE, 'default_style'), # STYLE ME
                 (stc.STC_PL_WORD, 'keyword_style') ]

if wx.VERSION >= (2, 9, 0, 0, ''):
    SYNTAX_ITEMS.append((stc.STC_PL_FORMAT, 'default_style')) #TODO
    SYNTAX_ITEMS.append((stc.STC_PL_FORMAT_IDENT, 'default_style')) #TODO
    SYNTAX_ITEMS.append((stc.STC_PL_SUB_PROTOTYPE, 'default_style')) #TODO

#---- Extra Properties ----#
FOLD = ("fold", "1")
FLD_COMPACT = ("fold.compact", "1")
FLD_COMMENT = ("fold.comment", "1")
FLD_POD = ("fold.perl.pod", "1")
FLD_PKG = ("fold.perl.package", "1")

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Perl""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_PERL)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [PERL_KW]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'#']

#---- Syntax Modules Internal Functions ----#
def KeywordString(option=0):
    """Returns the specified Keyword String
    @note: not used by most modules

    """
    if option == synglob.ID_LANG_PERL:
        return PERL_KW[1]
    else:
        return u''

#---- End Syntax Modules Internal Functions ----#

#-----------------------------------------------------------------------------#



###############################################################################
# Name: ed_fmgr.py                                                            #
# Purpose: Editra's Main Window Frame Manager                                 #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2011 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""


"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id:  $"
__revision__ = "$Revision:  $"

#--------------------------------------------------------------------------#
# Dependencies
import wx
#import wx.lib.agw.aui as aui
import wx.aui as aui

#--------------------------------------------------------------------------#
# Events

EVT_AUI_PANE_CLOSE = aui.EVT_AUI_PANE_CLOSE
EVT_AUI_PANE_RESTORE = aui.EVT_AUI_PANE_RESTORE

#--------------------------------------------------------------------------#

class EdFrameManager(aui.AuiManager):
    """Frame manager for external components to abstract underlying manager"""
    def __init__(self, wnd=None, flags=0):
        """Create the frame manager object
        @param wnd: Frame to manage
        @param flags: frame manager flags

        """
        super(EdFrameManager, self).__init__(wnd, flags)
        self.Bind(aui.EVT_AUI_PANE_CLOSE, self.OnPaneClose)
        self.Bind(aui.EVT_AUI_PANE_RESTORE, self.OnPaneRestore)

    def OnPaneClose(self, evt):
        """Notify pane it is being closed"""
#        pane = evt.GetPane()
#        print "Close", pane.window
#        nevt = wx.ShowEvent(pane.window.Id, False)
#        wx.PostEvent(pane.window, nevt)
        evt.Skip()

    def OnPaneRestore(self, evt):
        """Notify pane it is being opened"""
#        pane = evt.GetPane()
#        print "Restore:", pane.window
#        nevt = wx.ShowEvent(pane.window.Id, True)
#        wx.PostEvent(pane.window, nevt)
        evt.Skip()

    def AddPane(self, wnd, info=None, caption=None):
        return super(EdFrameManager, self).AddPane(wnd, info, caption)

    def IsEditorMaximized(self):
        """Is the editor pane maximized?
        return: bool

        """
        bEditMax = True
        # If any other pane is open then the editor is not maximized
        for pane in self.GetAllPanes():
            if pane.IsShown() and pane.name != "EditPane":
                bEditMax = False
                break
        return bEditMax

#--------------------------------------------------------------------------#

EdPaneInfo = aui.AuiPaneInfo
#    """Frame manager panel info"""
#    def __init__(self):
#        super(EdPaneInfo, self).__init__()



###############################################################################
# Name: ControlBoxDemo.py                                                     #
# Purpose: ControlBox and ControlBar Test and Demo File                       #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# Licence: wxWindows Licence                                                  #
###############################################################################

"""
Test file for testing the ControlBox, ControlBar, and SegmentBar classes

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: ControlBoxDemo.py 66765 2011-01-26 04:31:28Z CJP $"
__revision__ = "$Revision: 66765 $"

#-----------------------------------------------------------------------------#
# Imports
import time
import random
import sys
import os
import wx

#sys.path.insert(0, os.path.abspath('../../src'))
import eclib

# Local imports
import IconFile

#-----------------------------------------------------------------------------#
# Globals
ID_SHOW_CONTROL = wx.NewId()
ID_SHOW_SEGMENT = wx.NewId()

#-----------------------------------------------------------------------------#

class TestPanel(eclib.ControlBox):
    def __init__(self, parent, log):
        super(TestPanel, self).__init__(parent)

        # Attributes
        self.log = log

        # Setup
        self.CreateControlBar()
        cbar = self.GetControlBar()
        cbar.SetMargins(2, 2)
        cbar.AddControl(wx.Button(cbar, ID_SHOW_CONTROL, label="Show ControlBar Sample"))
        cbar.AddControl(wx.Button(cbar, ID_SHOW_SEGMENT, label="Show SegmentBar Sample"))
        text = wx.TextCtrl(self, style=wx.TE_MULTILINE|wx.TE_RICH2)
        text.SetValue("Welcome to the ControlBox Sample.\n\nThis window is a "
                      "ControlBox containing a TextCtrl and a ControlBar in it.\n\n"
                      "Click a button to see the extended samples.")
        self.SetWindow(text)

        # Event Handlers
        self.Bind(wx.EVT_BUTTON, self.OnButton)

    def OnButton(self, evt):
        """Handle Button Events"""
        e_id = evt.GetId()
        if e_id == ID_SHOW_CONTROL:
            # Show a Frame with a ControlBox using ControlBars in it
            frame = MakeTestFrame(self, "ControlBar Sample", self.log, False)
            frame.Show()
        elif e_id == ID_SHOW_SEGMENT:
            # Show a frame with a ControlBox using SegmentBars in it
            frame = MakeTestFrame(self, "SegmentBar Sample", self.log, True)
            frame.Show()
        else:
            evt.Skip()

#-----------------------------------------------------------------------------#

class ControlBarPanel(eclib.ControlBox):
    def __init__(self, parent, log):
        super(ControlBarPanel, self).__init__(parent)

        # Attributes
        self.log = log
        self.gauge = None
        self._timer = wx.Timer(self)

        # Setup
        if random.choice((True, False)):
            orient1 = wx.TOP
            orient2 = wx.BOTTOM
            edgealign = wx.ALIGN_RIGHT
        else:
            orient1 = wx.LEFT
            orient2 = wx.RIGHT
            edgealign = wx.ALIGN_BOTTOM

        cbar = self.CreateControlBar(pos=orient1)
        cbar.SetMargins(1, 1)
        err_bmp = wx.ArtProvider.GetBitmap(wx.ART_ERROR, wx.ART_MENU, (16, 16))
        w_bmp = wx.ArtProvider.GetBitmap(wx.ART_WARNING, wx.ART_MENU, (16, 16))
        cbar.AddTool(wx.ID_ANY, err_bmp, "hello world")
        cbar.AddTool(wx.ID_ANY, w_bmp, "warning")
        # only add other controls in horizontal mode (for aesthetics reasons)
        if not cbar.IsVerticalMode():
            cbar.AddStretchSpacer()
            choice = wx.Choice(cbar, wx.ID_ANY, choices=[str(x) for x in range(10)])
            cbar.AddControl(choice, edgealign)
            cbar.AddControl(wx.Button(cbar, label="New Frame"), edgealign)

        bbar = self.CreateControlBar(orient2)
        bbar.SetMargins(1, 1)
        bbar.AddTool(wx.ID_ANY, err_bmp, "HELLO")
        # Only add other controls in horizontal mode
        self.gauge = None
        if not bbar.IsVerticalMode():
            bbar.AddStretchSpacer()
            self.gauge = wx.Gauge(bbar, size=(100, 16))
            bbar.AddControl(self.gauge, edgealign)

        self.SetWindow(wx.TextCtrl(self, style=wx.TE_MULTILINE))
        self.Bind(eclib.EVT_CTRLBAR, self.OnControlBar)
        self.Bind(wx.EVT_BUTTON, self.OnButton)
        self.Bind(wx.EVT_TIMER, self.OnTimer)

        self._timer.Start(150)

    def OnControlBar(self, evt):
        self.log.write("ControlBarEvent: %d" % evt.GetId())

    def OnButton(self, evt):
        self.log.write("Button tool clicked: Id=%d" % evt.GetId())
        frame = MakeTestFrame(self, "Random Test Frame", self.log,
                              bool(long(time.time()) % 2))
        frame.Show()

    def OnTimer(self, evt):
        if self.gauge:
            self.gauge.Pulse()

#-----------------------------------------------------------------------------#

class SegmentPanel(eclib.ControlBox):
    def __init__(self, parent, log):
        super(SegmentPanel, self).__init__(parent)

        # Attributes
        self.log = log

        # Make up the top segment bar
        segbar = eclib.SegmentBar(self, style=eclib.CTRLBAR_STYLE_GRADIENT|\
                                              eclib.CTRLBAR_STYLE_LABELS|\
                                              eclib.CTRLBAR_STYLE_NO_DIVIDERS)
        for num in range(5):
            if num % 2:
                segbar.AddSegment(wx.NewId(), IconFile.Home.GetBitmap(), label=u'Home')
            else:
                segbar.AddSegment(wx.NewId(), IconFile.Monkey.GetBitmap(), label=u'Monkey')

        segbar.SetSegmentOption(4, eclib.SEGBTN_OPT_CLOSEBTNR)
        segbar.SetSegmentOption(3, eclib.SEGBTN_OPT_CLOSEBTNL)
        segbar.SetSegmentOption(2, eclib.SEGBTN_OPT_CLOSEBTNL)

        # Make a bottom segment bar
        segbar2 = eclib.SegmentBar(self, style=eclib.CTRLBAR_STYLE_GRADIENT)
        for num in range(5):
            if num % 2:
                segbar2.AddSegment(wx.NewId(), IconFile.Book.GetBitmap())
            else:
                segbar2.AddSegment(wx.NewId(), IconFile.Address.GetBitmap())

        # Put the components together
        self.SetControlBar(segbar, wx.TOP)
        self.SetControlBar(segbar2, wx.BOTTOM)
        self.SetWindow(wx.TextCtrl(self, style=wx.TE_MULTILINE, value="Hello World"))

        # Event Handlers
        self.Bind(eclib.EVT_SEGMENT_SELECTED, self.OnSegmentClicked)

    def OnSegmentClicked(self, evt):
        cur = evt.GetCurrentSelection()
        pre = evt.GetPreviousSelection()
        self.log.write("Segment Clicked: Cur=%d, Pre=%d, Id=%d" % (cur, pre, evt.GetId()))

#-----------------------------------------------------------------------------#

def MakeTestFrame(caller, title, log, segment=False):
    frame = wx.Frame(None, title=title)
    fsizer = wx.BoxSizer(wx.VERTICAL)
    if not segment:
        panel = ControlBarPanel(frame, log)
    else:
        panel = SegmentPanel(frame, log)
    fsizer.Add(panel, 1, wx.EXPAND)
    frame.SetSizer(fsizer)

    # Adjust Window Position
    if caller is not None:
        pos = caller.GetScreenPosition()
        frame.SetPosition((pos[0]+22, pos[1]+22))

    return frame

#-----------------------------------------------------------------------------#

class TestLog:
    def __init__(self):
        pass

    def write(self, msg):
        print msg

#-----------------------------------------------------------------------------#

overview = eclib.ctrlbox.__doc__
title = "ControlBox"

#-----------------------------------------------------------------------------#

if __name__ == '__main__':
    try:
        import run
    except ImportError:
        app = wx.PySimpleApp(False)
        frame = wx.Frame(None, title="ControlBox Test")
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(TestPanel(frame, TestLog()), 1, wx.EXPAND)
        frame.SetSizer(sizer)
        frame.Show()
        app.MainLoop()
    else:
        run.main(['', os.path.basename(sys.argv[0])] + sys.argv[1:])
